import numpy as np

from pySPFM.deconvolution import stability_selection


def test_get_subsampling_indices():
    n_scans = 10
    n_echos = 3

    # Test same mode
    mode = "same"
    subsample_result = np.array([1, 3, 5, 6, 7, 8])
    np.random.seed(200)
    subsample_idx = stability_selection.get_subsampling_indices(n_scans, n_echos, mode)
    assert np.allclose(subsample_idx, subsample_result)

    # Test different mode
    mode = "different"
    subsample_result = np.array([1, 3, 5, 6, 7, 8, 10, 12, 14, 15, 17, 19, 21, 22, 23, 24, 28, 29])
    np.random.seed(200)
    subsample_idx = stability_selection.get_subsampling_indices(n_scans, n_echos, mode)
    assert np.allclose(
        subsample_idx,
        subsample_result,
    )


def test_calculate_auc():
    n_surrogates = 10
    n_lambdas = 10
    lambdas = np.random.rand(n_lambdas * n_surrogates)
    coefs = np.random.rand(n_lambdas, n_surrogates)

    # Test if auc is calculated correctly
    auc = np.round(stability_selection.calculate_auc(coefs, lambdas, n_surrogates))

    assert auc <= 1.0
    assert auc >= 0.0


def test_stability_selection(sim_data, sim_hrf):
    y = np.load(sim_data, allow_pickle=True)[:, 0]
    hrf_matrix = np.load(sim_hrf, allow_pickle=True)
    n_lambdas = 20
    n_surrogates = 20

    # Test if stability selection works
    auc = stability_selection.stability_selection(hrf_matrix, y, n_lambdas, n_surrogates)
    assert auc.shape[0] == hrf_matrix.shape[1]
    assert np.count_nonzero(auc) > 0
    assert np.count_nonzero(auc) / auc.shape[0] < 0.5



"""HRF Matrix file."""
import logging
import subprocess

import numpy as np
from nilearn.glm.first_level import glover_hrf, spm_hrf

LGR = logging.getLogger("GENERAL")
RefLGR = logging.getLogger("REFERENCES")


class HRFMatrix:
    """A class for generating an HRF matrix.

    Parameters
    ----------
    TE : list
        Values of TE in ms, by default None
    model : str
        Model to use for HRF, by default "spm"
    block : bool
        Whether to use the block model in favor of the spike model, by default false
    custom : str
        Path to custom HRF file, by default None
    """

    def __init__(
        self,
        te=None,
        model="spm",
        block=True,
    ):
        # If te values are higher than 1, then assume they are in ms and convert to s
        # If te values are lower than 1, then assume they are in s

        if te is not None:
            if max(te) > 1:
                te = [i / 1000 for i in te]
            self.te = te

        self.model = model
        self.block = block

    def generate_hrf(self, tr, n_scans):
        """Generate HRF matrix.

        Parameters
        ----------
        tr : float
            tr of the acquisition.
        n_scans : int
            Number of scans.

        Returns
        -------
        self.hrf_ : array_like
            A hemodynamic response function (HRF).
        """

        # Read custom HRF from file if self.model ends in .1D or .txt
        if self.model.endswith(".1D") or self.model.endswith(".txt"):
            hrf = np.loadtxt(self.custom)

            # Make sure that the HRF is not longer than the number of scans
            # If it is, then raise an error
            if len(hrf) > n_scans:
                raise ValueError(
                    "HRF is longer than the number of scans. "
                    "Please make sure that your custom HRF is not longer than the number of scans."
                )
        elif self.model == "spm":
            hrf = spm_hrf(tr, oversampling=1, time_length=n_scans * tr)
        elif self.model == "glover":
            hrf = glover_hrf(tr, oversampling=1, time_length=n_scans * tr)
        else:
            raise ValueError(
                f"Model must be either 'spm', 'glover' or a custom '.1D' or '.txt' file, "
                f"not {self.model}"
            )

        # Generate HRF matrix
        hrf_mtx = hrf
        for i in range(n_scans - 1):
            foo = np.append(np.zeros(i + 1), hrf[: len(hrf) - i - 1])
            hrf_mtx = np.column_stack((hrf_mtx, foo))

        # Concatenate and scale HRFs for multi-echo,
        # leave it as it is for single-echo.
        if len(self.te) > 1:
            # Add integrator if necessary
            hrf_mtx_te = (
                -self.te[0] * np.dot(hrf_mtx, np.tril(np.ones(n_scans)))
                if self.block
                else -self.te[0] * hrf_mtx
            )

            # Calculate maximum HRF value
            max_val = np.max(abs(hrf_mtx_te))

            # Concatenate and scale HRFs for multi-echo
            for teidx in range(len(self.te) - 1):
                # Add integrator if necessary
                if self.block:
                    hrf_mtx_te = np.vstack(
                        (
                            hrf_mtx_te,
                            -self.te[teidx + 1] * np.dot(hrf_mtx, np.tril(np.ones(n_scans))),
                        )
                    )
                else:
                    hrf_mtx_te = np.vstack((hrf_mtx_te, -self.te[teidx + 1] * hrf_mtx))

            # Normalize HRF
            self.hrf_ = hrf_mtx_te / max_val
        elif self.block:
            hrf_non_norm = np.dot(hrf_mtx, np.tril(np.ones(n_scans)))

            # Normalize HRF
            self.hrf_ = hrf_non_norm / np.max(abs(hrf_non_norm))
        else:
            self.hrf_ = hrf_mtx / np.max(abs(hrf_mtx))

        return self

    def hrf_afni(self, hrf_model="SPMG1"):
        """Generate HRF with AFNI's 3dDeconvolve.

        Parameters
        ----------
        tr : float
            tr of the acquisition.
        hrf_model : str
            3dDeconvolve option to select HRF shape, by default "SPMG1"

        Returns
        -------
        hrf : array_like
            A hemodynamic response function (HRF).

        Notes
        -----
        AFNI installation is needed as it runs 3dDeconvolve on the terminal with subprocess.
        """
        dur_hrf = 8
        last_hrf_sample = 1
        # Increases duration until last HRF sample is zero
        while last_hrf_sample != 0:
            dur_hrf = 2 * dur_hrf
            hrf_command = (
                "3dDeconvolve -x1D_stop -nodata %d %f -polort -1 -num_stimts 1 -stim_times 1 "
                "'1D:0' '%s' -quiet -x1D stdout: | 1deval -a stdin: -expr 'a'"
                % (dur_hrf, self, hrf_model)
            )
            hrf_tr_str = subprocess.check_output(
                hrf_command, shell=True, universal_newlines=True
            ).splitlines()
            hrf = np.array([float(i) for i in hrf_tr_str])
            last_hrf_sample = hrf[len(hrf) - 1]
            if last_hrf_sample != 0:
                LGR.info(
                    "Duration of HRF was not sufficient for specified model. Doubling duration "
                    "and computing again."
                )

        # Removes tail of zero samples
        while last_hrf_sample == 0:
            hrf = hrf[:-1]
            last_hrf_sample = hrf[len(hrf) - 1]

        return hrf



"""Utility functions for aroma's test suite."""
import os.path as op


def get_tests_resource_path():
    """Return the path to general resources, terminated with separator.
    Resources are kept outside package folder in "datasets".
    Based on function by Yaroslav Halchenko used in Neurosynth Python package.
    """
    return op.abspath(op.join(op.dirname(__file__), "data") + op.sep)



import numpy as np

from pySPFM.deconvolution import hrf_generator


def test_HRF_matrix(spm_single_echo, spm_single_echo_block, glover_multi_echo):
    hrf_object = hrf_generator.HRFMatrix(te=[0], block=False)
    hrf = hrf_object.generate_hrf(tr=1, n_scans=168).hrf_
    hrf_loaded = np.load(spm_single_echo)
    assert np.allclose(hrf, hrf_loaded)

    hrf_object = hrf_generator.HRFMatrix(te=[0], block=True)
    hrf = hrf_object.generate_hrf(tr=1, n_scans=168).hrf_
    hrf_loaded = np.load(spm_single_echo_block)
    assert np.allclose(hrf, hrf_loaded)

    te = np.array([18.4, 32.0, 48.4])
    hrf_object = hrf_generator.HRFMatrix(te=te, block=False, model="glover")
    hrf = hrf_object.generate_hrf(tr=1, n_scans=168).hrf_
    hrf_loaded = np.load(glover_multi_echo)
    assert np.allclose(hrf, hrf_loaded)



"""
Integration tests for "real" data
"""

import glob
import os
import re
import shutil
import tarfile
from gzip import GzipFile
from io import BytesIO

import pytest
import requests
from pkg_resources import resource_filename


def download_test_data(osf, outpath):
    """
    Downloads tar.gz data stored at `osf` and unpacks into `outpath` (taken from tedana)
    Parameters
    ----------
    osf : str
        URL to OSF file that contains data to be downloaded
    outpath : str
        Path to directory where OSF data should be extracted
    """

    req = requests.get(osf)
    req.raise_for_status()
    t = tarfile.open(fileobj=GzipFile(fileobj=BytesIO(req.content)))
    os.makedirs(outpath, exist_ok=True)
    t.extractall(outpath)


def check_integration_outputs(fname, outpath, workflow="pySPFM"):
    """
    Checks outputs of integration tests (taken from tedana)
    Parameters
    ----------
    fname : str
        Path to file with expected outputs
    outpath : str
        Path to output directory generated from integration tests
    """

    # Gets filepaths generated by integration test
    existing = [
        os.path.relpath(f, outpath)
        for f in glob.glob(os.path.join(outpath, "**"), recursive=True)[1:]
    ]

    # Checks for log file
    if workflow == "pySPFM":
        log_regex = "^pySPFM_[12][0-9]{3}-[0-9]{2}-[0-9]{2}T[0-9]{2}[0-9]{2}[0-9]{2}.tsv$"
    elif workflow == "auc_to_estimates":
        log_regex = (
            "^auc_to_estimates_[12][0-9]{3}-[0-9]{2}-[0-9]{2}T[0-9]{2}[0-9]{2}[0-9]{2}.tsv$"
        )
    logfiles = [out for out in existing if re.match(log_regex, out)]
    assert len(logfiles) == 1

    # Removes logfile from list of existing files
    existing.remove(logfiles[0])

    # Compares remaining files with those expected
    with open(fname, "r") as f:
        tocheck = f.read().splitlines()
    tocheck = [os.path.normpath(path) for path in tocheck]
    assert sorted(tocheck) == sorted(existing)


def test_integration_five_echo(skip_integration, script_runner, mask_five_echo):
    """Integration test of the full pySPFM workflow using five-echo test data."""

    if skip_integration:
        pytest.skip("Skipping five-echo integration test")

    out_dir = "/tmp/data/five-echo/pySPFM.five-echo"

    if os.path.exists(out_dir):
        shutil.rmtree(out_dir)

    # download data and run the test
    download_test_data("https://osf.io/vg4wy/download", os.path.dirname(out_dir))
    prepend = "/tmp/data/five-echo/p06.SBJ01_S09_Task11_e"
    suffix = ".psc.nii.gz"
    datalist = [prepend + str(i + 1) + suffix for i in range(5)]
    echo_times = [15.4, 29.7, 44.0, 58.3, 72.6]

    # CLI args
    args = (
        ["pySPFM", "-i"]
        + datalist
        + ["-te"]
        + [str(te) for te in echo_times]
        + ["-m"]
        + [mask_five_echo]
        + ["-o"]
        + ["test-me"]
        + ["-tr"]
        + ["2"]
        + ["-d"]
        + [out_dir]
        + ["-crit"]
        + ["factor"]
        + ["-factor"]
        + ["10"]
        + ["--max_iter_fista"]
        + ["50"]
        + ["-j"]
        + ["1"]
        + [
            "--debug",
            "--debias",
            "--block",
            "--bids",
        ]
    )
    ret = script_runner.run(*args)
    assert ret.success

    # compare the generated output files
    fn = resource_filename("pySPFM", "tests/data/nih_five_echo_outputs_verbose.txt")
    check_integration_outputs(fn, out_dir)


def test_integration_lars(skip_integration, script_runner, mask_five_echo):
    """Integration test of the full pySPFM workflow using five-echo test data."""

    if skip_integration:
        pytest.skip("Skipping five-echo integration test")

    out_dir = "/tmp/data/five-echo/pySPFM.five-echo"

    if os.path.exists(out_dir):
        shutil.rmtree(out_dir)

    # download data and run the test using the second echo time
    download_test_data("https://osf.io/vg4wy/download", os.path.dirname(out_dir))
    prepend = "/tmp/data/five-echo/p06.SBJ01_S09_Task11_e"
    suffix = ".psc.nii.gz"
    data = f"{prepend}2{suffix}"

    # CLI args
    args = (
        ["pySPFM", "-i"]
        + [data]
        + ["-m"]
        + [mask_five_echo]
        + ["-o"]
        + ["test_lars"]
        + ["-tr"]
        + ["2"]
        + ["-d"]
        + [out_dir]
        + ["-crit"]
        + ["bic"]
        + ["--max_iter_factor"]
        + ["0.3"]
        + ["-j"]
        + ["1"]
        + [
            "--debug",
            "--debias",
        ]
    )
    ret = script_runner.run(*args)
    assert ret.success

    # compare the generated output files
    fn = resource_filename("pySPFM", "tests/data/lars_integration_outputs.txt")
    check_integration_outputs(fn, out_dir)


def test_integration_stability_selection(skip_integration, script_runner, mask_five_echo):
    """Integration test of the pySPFM stability selection workflow using five-echo test data."""

    if skip_integration:
        pytest.skip("Skipping five-echo integration test")

    out_dir = "/tmp/data/five-echo/pySPFM.five-echo"

    if os.path.exists(out_dir):
        shutil.rmtree(out_dir)

    # download data and run the test using the second echo time
    download_test_data("https://osf.io/vg4wy/download", os.path.dirname(out_dir))
    prepend = "/tmp/data/five-echo/p06.SBJ01_S09_Task11_e"
    suffix = ".psc.nii.gz"
    datalist = [prepend + str(i + 1) + suffix for i in range(5)]
    echo_times = [15.4, 29.7, 44.0, 58.3, 72.6]

    # CLI args
    args = (
        ["pySPFM", "-i"]
        + datalist
        + ["-te"]
        + [str(te) for te in echo_times]
        + ["-m"]
        + [mask_five_echo]
        + ["-o"]
        + ["test_stability"]
        + ["-tr"]
        + ["2"]
        + ["-d"]
        + [out_dir]
        + ["-crit"]
        + ["stability"]
        + ["--max_iter_factor"]
        + ["0.3"]
        + ["-j"]
        + ["1"]
        + [
            "--debug",
            "--debias",
        ]
    )

    ret = script_runner.run(*args)
    assert ret.success

    # compare the generated output files
    fn = resource_filename("pySPFM", "tests/data/stability_integration_outputs.txt")
    check_integration_outputs(fn, out_dir)


def test_integration_auc_to_estimates(
    skip_integration, script_runner, mask_five_echo, test_AUC, mean_AUC, auc_4D_thr
):
    if skip_integration:
        pytest.skip("Skipping five-echo integration test")

    out_dir = "/tmp/data/five-echo/pySPFM.five-echo"

    if os.path.exists(out_dir):
        shutil.rmtree(out_dir)

    # download data and run the test using the second echo time
    download_test_data("https://osf.io/vg4wy/download", os.path.dirname(out_dir))
    prepend = "/tmp/data/five-echo/p06.SBJ01_S09_Task11_e"
    suffix = ".psc.nii.gz"
    datalist = [prepend + str(i + 1) + suffix for i in range(5)]
    echo_times = [15.4, 29.7, 44.0, 58.3, 72.6]

    ############################
    # CLI args for 95th percentile static thresholding
    args = (
        ["auc_to_estimates", "-i"]
        + datalist
        + ["-te"]
        + [str(te) for te in echo_times]
        + ["-m"]
        + [mask_five_echo, mask_five_echo]
        + ["-a"]
        + [test_AUC]
        + ["-thr"]
        + ["95"]
        + ["-o"]
        + ["test_auc2est"]
        + ["-tr"]
        + ["2"]
        + ["-d"]
        + [out_dir]
        + ["-j"]
        + ["1"]
        + [
            "--debug",
        ]
    )

    ret = script_runner.run(*args)
    assert ret.success

    # compare the generated output files
    fn = resource_filename("pySPFM", "tests/data/auc_to_estimates_outputs.txt")
    check_integration_outputs(fn, out_dir, "auc_to_estimates")

    ############################
    if os.path.exists(out_dir):
        shutil.rmtree(out_dir)
    # CLI args for 95th percentile dynamic thresholding
    args = (
        ["auc_to_estimates", "-i"]
        + datalist
        + ["-te"]
        + [str(te) for te in echo_times]
        + ["-m"]
        + [mask_five_echo, mask_five_echo]
        + ["-a"]
        + [test_AUC]
        + ["-thr"]
        + ["95"]
        + ["--strategy"]
        + ["time"]
        + ["-o"]
        + ["test_auc2est"]
        + ["-tr"]
        + ["2"]
        + ["-d"]
        + [out_dir]
        + ["-j"]
        + ["1"]
        + [
            "--debug",
        ]
    )

    ret = script_runner.run(*args)
    assert ret.success

    # compare the generated output files
    fn = resource_filename("pySPFM", "tests/data/auc_to_estimates_outputs.txt")
    check_integration_outputs(fn, out_dir, "auc_to_estimates")

    ############################
    if os.path.exists(out_dir):
        shutil.rmtree(out_dir)
    # CLI args for 3D mask thresholding
    args = (
        ["auc_to_estimates", "-i"]
        + datalist
        + ["-te"]
        + [str(te) for te in echo_times]
        + ["-m"]
        + [mask_five_echo, mean_AUC]
        + ["-a"]
        + [test_AUC]
        + ["-o"]
        + ["test_auc2est"]
        + ["-tr"]
        + ["2"]
        + ["-d"]
        + [out_dir]
        + ["-j"]
        + ["1"]
        + [
            "--debug",
        ]
    )

    ret = script_runner.run(*args)
    assert ret.success

    # compare the generated output files
    fn = resource_filename("pySPFM", "tests/data/auc_to_estimates_outputs.txt")
    check_integration_outputs(fn, out_dir, "auc_to_estimates")

    ############################
    if os.path.exists(out_dir):
        shutil.rmtree(out_dir)
    # CLI args for 4D mask thresholding
    args = (
        ["auc_to_estimates", "-i"]
        + datalist
        + ["-te"]
        + [str(te) for te in echo_times]
        + ["-m"]
        + [mask_five_echo, auc_4D_thr]
        + ["-a"]
        + [test_AUC]
        + ["-o"]
        + ["test_auc2est"]
        + ["-tr"]
        + ["2"]
        + ["-d"]
        + [out_dir]
        + ["-j"]
        + ["1"]
        + [
            "--debug",
        ]
    )

    ret = script_runner.run(*args)
    assert ret.success

    # compare the generated output files
    fn = resource_filename("pySPFM", "tests/data/auc_to_estimates_outputs.txt")
    check_integration_outputs(fn, out_dir, "auc_to_estimates")



import argparse
import datetime
import logging
import os
import sys
from os import path as op

import numpy as np
from dask import compute
from dask import delayed as delayed_dask

from pySPFM import __version__, utils
from pySPFM.deconvolution import (
    debiasing,
    fista,
    hrf_generator,
    lars,
    select_lambda,
    spatial_regularization,
    stability_selection,
)
from pySPFM.io import read_data, write_data, write_json
from pySPFM.utils import dask_scheduler, get_outname
from pySPFM.workflows.parser_utils import check_hrf_value, is_valid_file

LGR = logging.getLogger("GENERAL")
RefLGR = logging.getLogger("REFERENCES")


def _get_parser():
    """
    Parse command line inputs for this function.

    Returns
    -------
    parser.parse_args() : argparse dict

    Notes
    -----
    # Argument parser follow template provided by RalphyZ.
    # https://stackoverflow.com/a/43456577
    """
    parser = argparse.ArgumentParser()
    optional = parser._action_groups.pop()
    required = parser.add_argument_group("Required Argument:")
    required.add_argument(
        "-i",
        "--input",
        dest="data_fn",
        type=lambda x: is_valid_file(parser, x),
        nargs="+",
        help="The name of the file containing fMRI data. ",
        required=True,
    )
    required.add_argument(
        "-m",
        "--mask",
        dest="mask_fn",
        type=lambda x: is_valid_file(parser, x),
        help="The name of the file containing the mask for the fMRI data. ",
        required=True,
    )
    required.add_argument(
        "-o",
        "--output",
        dest="output_filename",
        type=str,
        help="The name of the output file with no extension.",
        required=True,
    )
    required.add_argument(
        "-tr",
        dest="tr",
        type=float,
        help="TR of the fMRI data acquisition.",
        required=True,
    )
    optional.add_argument(
        "-d",
        "--dir",
        dest="out_dir",
        type=str,
        help="Output directory. Default is current.",
        default=".",
    )
    optional.add_argument(
        "-te",
        dest="te",
        nargs="*",
        type=float,
        help="List with TE of the fMRI data acquisition.",
        default=[0],
    )
    optional.add_argument(
        "-hrf",
        "--hrf",
        dest="hrf_model",
        type=check_hrf_value,
        help=(
            "HRF model to use. Default is 'spm'. Options are 'spm', 'glover', or a custom HRF "
            "file with the '.1D' or '.txt' extension."
        ),
        default="spm",
    )
    optional.add_argument(
        "-block",
        "--block",
        dest="block_model",
        action="store_true",
        help="Estimate innovation signals. Default = False.",
        default=False,
    )
    optional.add_argument(
        "--debias",
        dest="debias",
        action="store_true",
        help="Perform debiasing step. Default = False.",
        default=False,
    )
    optional.add_argument(
        "-g",
        "--group",
        dest="group",
        type=float,
        help="Weight of the grouping in space (we suggest not going "
        "higher than 0.3). Default = 0.",
        default=0,
    )
    optional.add_argument(
        "-crit",
        "--criterion",
        dest="criterion",
        type=str,
        choices=[
            "mad",
            "mad_update",
            "ut",
            "lut",
            "factor",
            "pcg",
            "eigval",
            "bic",
            "aic",
            "stability",
        ],
        help="Criteria with which lambda is selected to estimate activity-inducing "
        "and innovation signals. 'stability' performs the stability selection technique."
        "'aic' and 'bic' are used with the LARS algorithm, "
        " while the other criteria are used with FISTA. Default = 'ut'.",
        default="ut",
    )
    optional.add_argument(
        "-pcg",
        "--percentage",
        dest="pcg",
        type=float,
        help=(
            "Percentage of maximum lambda to use on temporal regularization with FISTA "
            "(default = None)."
        ),
        default=None,
    )
    optional.add_argument(
        "-factor",
        "--factor",
        dest="factor",
        type=float,
        help="Factor to multiply noise estimate for lambda selection.",
        default=1,
    )
    optional.add_argument(
        "-lambda_echo",
        "--lambda_echo",
        dest="lambda_echo",
        type=int,
        help="Number of the TE data to calculate lambda for (default = last TE).",
        default=-1,
    )
    optional.add_argument(
        "--max_iter_factor",
        dest="max_iter_factor",
        type=float,
        help=(
            "Factor of number of samples to use as the maximum number of iterations for LARS "
            "(default = 1.0)."
        ),
        default=1.0,
    )
    optional.add_argument(
        "--max_iter_fista",
        dest="max_iter_fista",
        type=int,
        help="Maximum number of iterations for FISTA (default = 400).",
        default=400,
    )
    optional.add_argument(
        "--min_iter_fista",
        dest="min_iter_fista",
        type=int,
        help="Minimum number of iterations for FISTA (default = 50).",
        default=50,
    )
    optional.add_argument(
        "--max_iter_spatial",
        dest="max_iter_spatial",
        type=int,
        help="Maximum number of iterations for spatial regularization (default = 100).",
        default=100,
    )
    optional.add_argument(
        "--max_iter",
        dest="max_iter",
        type=int,
        help=(
            "Maximum number of iterations for alternating temporal and spatial regularizations "
            "(default = 10)."
        ),
        default=10,
    )
    optional.add_argument(
        "--jobqueue",
        help="Jobqueue.yaml file to set up parallel processing (default = None).",
        default=None,
        type=str,
        dest="jobqueue",
    )
    optional.add_argument(
        "-j",
        "--jobs",
        dest="n_jobs",
        type=int,
        help="Number of jobs to parallelize for loops (default = 4).",
        default=4,
    )
    optional.add_argument(
        "-spatial",
        "--spatial_weight",
        dest="spatial_weight",
        type=float,
        help=(
            "Weight for spatial regularization estimates (estimates of temporal regularization "
            "are equal to 1 minus this value). A value of 0 means only temporal regularization "
            "is applied. Default=0"
        ),
        default=0,
    )
    optional.add_argument(
        "--spatial_lambda",
        dest="spatial_lambda",
        type=float,
        help="Lambda for spatial regularization. Default=1",
        default=1,
    )
    optional.add_argument(
        "--spatial_dim",
        dest="spatial_dim",
        type=int,
        choices=[2, 3],
        help=(
            "Slice-wise regularization with dim = 2; whole-volume regularization with dim=3. "
            "Default = 3."
        ),
        default=3,
    )
    optional.add_argument(
        "-mu",
        "--mu",
        dest="mu",
        type=float,
        help="Step size for spatial regularization (default = 0.01).",
        default=0.01,
    )
    optional.add_argument(
        "-tol",
        "--tolerance",
        dest="tolerance",
        type=float,
        help="Tolerance for FISTA (default = 1e-6).",
        default=1e-6,
    )
    optional.add_argument(
        "-bids",
        "--bids",
        dest="use_bids",
        action="store_true",
        help=(
            "Use BIDS-style suffix on the given `output` (default = False). pySPFM assumes that "
            "`output` follows the BIDS convention. Not using this option will default to using "
            "AFNI to update the header of the output."
        ),
        default=False,
    )
    optional.add_argument(
        "-nsur",
        "--nsurrogates",
        dest="n_surrogates",
        type=int,
        help="Number of surrogates to generate for stability selection (default = 50).",
        default=50,
    )
    optional.add_argument(
        "-debug",
        "--debug",
        dest="debug",
        action="store_true",
        help="Logs in the terminal will have increased "
        "verbosity, and will also be written into "
        "a .tsv file in the output directory.",
        default=False,
    )
    optional.add_argument(
        "-quiet",
        "--quiet",
        dest="quiet",
        help=argparse.SUPPRESS,
        action="store_true",
        default=False,
    )
    optional.add_argument("-v", "--version", action="version", version=("%(prog)s " + __version__))

    parser._action_groups.append(optional)

    return parser


def pySPFM(
    data_fn,
    mask_fn,
    output_filename,
    tr,
    out_dir=".",
    te=[0],
    hrf_model="spm",
    block_model=False,
    debias=True,
    group=0.2,
    criterion="bic",
    pcg=0.8,
    factor=10,
    lambda_echo=-1,
    max_iter_factor=1.0,
    max_iter_fista=500,
    max_iter_spatial=100,
    max_iter=10,
    min_iter_fista=50,
    jobqueue=None,
    n_jobs=4,
    spatial_weight=0,
    spatial_lambda=1,
    spatial_dim=3,
    mu=0.01,
    tolerance=1e-6,
    use_bids=False,
    n_surrogates=50,
    debug=False,
    quiet=False,
    command_str=None,
):
    """pySPFM is a Python programm for the Sparse Paradigm Free Mapping algorithm.

    Parameters
    ----------
    data_fn : str
        Input data file(s). List of files for multi-echo data
    mask_fn : str
        Input mask file
    output_filename : str
        Prefix of output files.
    tr : float
        Repetition time
    out_dir : str
        Output directory
    te : list, optional
        TE values in ms, by default [0], i.e. single-echo
    hrf_model : str, optional
        HRF model to use, by default 'spm', i.e. SPM's canonical HRF.
        Options are 'spm', 'glover' and a path to a 1D file with a custom HRF model.
    block_model : bool, optional
        Use the block model instead of using the spike model, by default False
    debias : bool, optional
        Perform debiasing step to recover true amplitude of estimates, by default True
    group : float, optional
        Grouping (l2,1-norm) regularization parameter, by default 0.2
    criterion : str, optional
        Criterion to select regularization parameter lambda, by default "bic"
    pcg : float, optional
        Percentage of the maximum lambda possible to use as lambda, by default 0.8
    factor : int, optional
        Factor of the estimate of the level of noise to use as lambda, by default 10
    lambda_echo : int, optional
        When using multi-echo data, the number of TE to use to estimate the level of the noise,
        by default -1
    max_iter_factor : float, optional
        Percentage of TRs to limit the LARS search to (less is faster), by default 1.0
    max_iter_fista : int, optional
        Maximum number of iterations for FISTA, by default 500
    max_iter_spatial : int, optional
        Maximum number of iterations for spatial regularization, by default 100
    max_iter : int, optional
        Maximum number of iterations to solve both temporal and spatial regularization,
        by default 10
    min_iter_fista : int, optional
        Minimum number of iterations for FISTA, by default 50
    jobqueue : str, optional
        Jobqueue to use for parallel processing, by default None
    n_jobs : int, optional
        Number of parallel jobs to use on for loops, by default 4
    spatial_weight : int, optional
        Weighting between the temporal and spatial regularization, by default 0 (only temporal)
    spatial_lambda : int, optional
        Regularization parameter for spatial regularization, by default 1
    spatial_dim : int, optional
        Dimensions of the spatial regularization filter (can be 2 for slices or 3 for volume),
        by default 3
    mu : float, optional
        Step size for spatial regularization, by default 0.01
    tolerance : float, optional
        Tolerance for residuals to find convergence of inverse problem, by default 1e-6
    is_atlas : bool, optional
        Read mask as atlas with different labels, by default False
    use_bids : bool, optional
        Use BIDS-style suffix on the given `output` (default = False). pySPFM assumes that `output`
        follows the BIDS convention. Not using this option will default to using AFNI to update the
        header of the output."
    n_surrogates : int, optional
        Number of surrogates to generate for stability selection, by default 50
    debug : bool, optional
        Logger option for debugging, by default False
    quiet : bool, optional
        Quiet logger option (no messages shown), by default False
    command_str : str, optional
        Command string to be used in the log file, by default None.

    Raises
    ------
    ValueError
        If wrong criterion is provided.
    """
    # Generate output directory if it doesn't exist
    out_dir = op.abspath(out_dir)
    if not op.isdir(out_dir):
        os.mkdir(out_dir)

    # Save command into sh file, if the command-line interface was used
    if command_str is not None:
        command_file = open(os.path.join(out_dir, "call.sh"), "w")
        command_file.write(command_str)
        command_file.close()

    LGR = logging.getLogger("GENERAL")
    # RefLGR = logging.getLogger("REFERENCES")
    # create logfile name
    basename = "pySPFM_"
    extension = "tsv"
    start_time = datetime.datetime.now().strftime("%Y-%m-%dT%H%M%S")
    logname = op.join(out_dir, (basename + start_time + "." + extension))
    refname = op.join(out_dir, "_references.txt")
    utils.setup_loggers(logname, refname, quiet=quiet, debug=debug)

    LGR.info("Using output directory: {}".format(out_dir))

    n_te = len(te)

    if all(i >= 1 for i in te):
        te = [x / 1000 for x in te]

    LGR.info("Reading data...")
    if n_te == 1:
        data_masked, masker = read_data(data_fn[0], mask_fn)
        n_scans = data_masked.shape[0]
        n_voxels = data_masked.shape[1]
    elif n_te > 1:
        # If the first element of data_fn has spaces, it is a list of paths
        # Convert it into a list
        if " " in data_fn[0]:
            data_fn = data_fn[0].split(" ")

        for te_idx in range(n_te):
            data_temp, masker = read_data(data_fn[te_idx], mask_fn)
            if te_idx == 0:
                data_masked = data_temp
                n_scans = data_temp.shape[0]
                n_voxels = data_temp.shape[1]
            else:
                # data_temp, _, _, _ = read_data(data_fn[te_idx], mask_fn, mask_idxs)
                data_masked = np.concatenate((data_masked, data_temp), axis=0)

            LGR.info(f"{te_idx + 1}/{n_te} echoes...")

    LGR.info("Data read.")

    # Generate design matrix with shifted versions of HRF
    LGR.info("Generating design matrix with shifted versions of HRF...")
    hrf_obj = hrf_generator.HRFMatrix(te=te, block=block_model, model=hrf_model)
    hrf = hrf_obj.generate_hrf(tr=tr, n_scans=n_scans).hrf_

    # Run LARS if bic or aic on given.
    # If another criterion is given, then solve with FISTA.
    lars_criteria = ["bic", "aic"]
    fista_criteria = ["mad", "mad_update", "ut", "lut", "factor", "pcg", "eigval"]

    # Run for loop only once for just temporal regularization
    if spatial_weight == 0:
        max_iter = 1
    else:
        # Initialize variables for spatial regularization
        estimates_temporal = np.empty((n_scans, n_voxels))
        estimates_spatial = np.empty((n_scans, n_voxels))
        final_estimates = np.empty((n_scans, n_voxels))

    # Initialize list to save keywords used for BIDS compatible outputs
    out_bids_keywords = []

    # Iterate between temporal and spatial regularizations
    client, _ = dask_scheduler(n_jobs, jobqueue)

    # Scatter data to workers if client is not None
    if client is not None:
        hrf_fut = client.scatter(hrf)
    else:
        hrf_fut = hrf

    # Solve stability selection
    if criterion == "stability":
        LGR.info("Solving inverse problem with stability selection...")
        n_lambdas = int(np.ceil(max_iter_factor * n_scans))
        auc = np.zeros((n_scans, n_voxels))

        # Solve stability regularization
        futures = [
            delayed_dask(stability_selection.stability_selection)(
                hrf_fut,
                data_masked[:, vox_idx],
                n_lambdas,
                n_surrogates,
            )
            for vox_idx in range(n_voxels)
        ]

        # Gather results
        if client is not None:
            stability_estimates = compute(futures)[0]
        else:
            stability_estimates = compute(futures, scheduler="single-threaded")[0]

        for vox_idx in range(n_voxels):
            auc[:, vox_idx] = np.squeeze(stability_estimates[vox_idx])

        LGR.info("Stability selection finished.")

        LGR.info("Saving AUCs to %s..." % out_dir)
        out_bids_keywords.append("AUC")
        output_name = get_outname(output_filename, "AUC", "nii.gz", use_bids)
        write_data(
            auc,
            os.path.join(out_dir, output_name),
            masker,
            data_fn[0],
            command_str,
            use_bids,
        )

    # Solve FISTA
    else:
        for iter_idx in range(max_iter):
            if spatial_weight > 0:
                data_temp_reg = final_estimates - estimates_temporal + data_masked
            else:
                data_temp_reg = data_masked

            estimates = np.zeros((n_scans, n_voxels))
            lambda_map = np.zeros(n_voxels)

            if criterion in lars_criteria:
                LGR.info("Solving inverse problem with LARS...")
                n_lambdas = int(np.ceil(max_iter_factor * n_scans))
                # Solve LARS for each voxel with parallelization
                futures = []
                for vox_idx in range(n_voxels):
                    fut = delayed_dask(lars.solve_regularization_path, pure=False)(
                        hrf_fut, data_temp_reg[:, vox_idx], n_lambdas, criterion
                    )
                    futures.append(fut)

                # Gather results
                if client is not None:
                    lars_estimates = compute(futures)[0]
                else:
                    lars_estimates = compute(futures, scheduler="single-threaded")[0]

                for vox_idx in range(n_voxels):
                    estimates[:, vox_idx] = np.squeeze(lars_estimates[vox_idx][0])
                    lambda_map[vox_idx] = np.squeeze(lars_estimates[vox_idx][1])

            elif criterion in fista_criteria:
                LGR.info("Solving inverse problem with FISTA...")
                # Solve fista
                futures = []
                for vox_idx in range(n_voxels):
                    fut = delayed_dask(fista.fista, pure=False)(
                        hrf_fut,
                        data_temp_reg[:, vox_idx],
                        criterion=criterion,
                        max_iter=max_iter_fista,
                        min_iter=min_iter_fista,
                        tol=tolerance,
                        group=group,
                        pcg=pcg,
                        factor=factor,
                        lambda_echo=lambda_echo,
                    )
                    futures.append(fut)

                # Gather results
                if client is not None:
                    fista_estimates = compute(futures)[0]
                else:
                    fista_estimates = compute(futures, scheduler="single-threaded")[0]

                for vox_idx in range(n_voxels):
                    estimates[:, vox_idx] = np.squeeze(fista_estimates[vox_idx][0])
                    lambda_map[vox_idx] = np.squeeze(fista_estimates[vox_idx][1])

            else:
                raise ValueError("Wrong criterion option given.")

            # Convolve with HRF
            if block_model:
                estimates_block = estimates
                hrf_obj = hrf_generator.HRFMatrix(te=te, block=False, model=hrf_model)
                hrf_fitting = hrf_obj.generate_hrf(tr=tr, n_scans=n_scans).hrf_
                estimates_spike = np.dot(np.tril(np.ones(n_scans)), estimates_block)
                fitts = np.dot(hrf_fitting, estimates_spike)
            else:
                estimates_spike = estimates
                fitts = np.dot(hrf, estimates_spike)

            # Perform spatial regularization if a weight is given
            if spatial_weight > 0:
                # Update temporal estimates
                estimates_temporal = estimates_temporal + (estimates - final_estimates)

                # Calculates for the whole volume
                estimates_tikhonov = spatial_regularization.spatial_tikhonov(
                    final_estimates,
                    final_estimates - estimates_spatial + data_masked,
                    masker,
                    max_iter_spatial,
                    spatial_dim,
                    spatial_lambda,
                    mu,
                )

                # Update spatial estimates
                estimates_spatial = estimates_spatial + (estimates_tikhonov - final_estimates)

                # Calculate final estimates
                final_estimates = (
                    estimates_temporal * (1 - spatial_weight) + spatial_weight * estimates_spatial
                )
            else:
                final_estimates = estimates

        LGR.info("Inverse problem solved.")

        # Update HRF for block model
        if block_model:
            hrf_obj = hrf_generator.HRFMatrix(te=te, block=False, model=hrf_model)
            hrf = hrf_obj.generate_hrf(tr=tr, n_scans=n_scans).hrf_

        # Perform debiasing step
        if debias:
            LGR.info("Debiasing estimates...")
            if block_model:
                estimates_spike = debiasing.debiasing_block(
                    hrf=hrf, y=data_masked, estimates_matrix=final_estimates
                )
                fitts = np.dot(hrf, estimates_spike)
            else:
                estimates_spike, fitts = debiasing.debiasing_spike(
                    hrf, data_masked, final_estimates
                )
        elif block_model:
            estimates_spike = np.dot(np.tril(np.ones(n_scans)), estimates_block)
            fitts = np.dot(hrf, estimates_spike)

        LGR.info("Saving results...")

        # Save innovation signal
        if block_model:
            estimates_block = final_estimates
            output_name = get_outname(output_filename, "innovation", "nii.gz", use_bids)
            out_bids_keywords.append("innovation")
            write_data(
                estimates_block,
                os.path.join(out_dir, output_name),
                masker,
                data_fn[0],
                command_str,
                use_bids=use_bids,
            )

            if not debias:
                hrf_obj = hrf_generator.HRFMatrix(TR=tr, n_scans=n_scans, TE=te, block=False)
                hrf = hrf_obj.generate_hrf().hrf
                estimates_spike = np.dot(np.tril(np.ones(n_scans)), estimates_block)
                fitts = np.dot(hrf, estimates_spike)

        # Save activity-inducing signal
        if n_te == 1:
            output_name = get_outname(output_filename, "activityInducing", "nii.gz", use_bids)
            out_bids_keywords.append("activityInducing")
        elif n_te > 1:
            output_name = get_outname(output_filename, "activityInducing", "nii.gz", use_bids)
            out_bids_keywords.append("activityInducing")
        write_data(
            estimates_spike,
            os.path.join(out_dir, output_name),
            masker,
            data_fn[0],
            command_str,
            use_bids=use_bids,
        )

        # Save fitts
        if n_te == 1:
            output_name = get_outname(output_filename, "denoised_bold", "nii.gz", use_bids)
            out_bids_keywords.append("denoised_bold")
            write_data(
                fitts,
                os.path.join(out_dir, output_name),
                masker,
                data_fn[0],
                command_str,
                use_bids=use_bids,
            )
        elif n_te > 1:
            for te_idx in range(n_te):
                te_data = fitts[te_idx * n_scans : (te_idx + 1) * n_scans, :]
                output_name = get_outname(
                    f"{output_filename}_echo-{te_idx + 1}", "denoised_bold", "nii.gz", use_bids
                )
                out_bids_keywords.append(f"echo-{te_idx + 1}_desc-denoised_bold")
                write_data(
                    te_data,
                    os.path.join(out_dir, output_name),
                    masker,
                    data_fn[0],
                    command_str,
                    use_bids=use_bids,
                )

        # Save noise estimate
        if n_te == 1:
            output_name = get_outname(output_filename, "MAD", "nii.gz", use_bids)
            out_bids_keywords.append("MAD")
            out_data = data_masked[:n_scans, :]
            _, _, noise_estimate = select_lambda.select_lambda(hrf=hrf, y=out_data)
            write_data(
                np.expand_dims(noise_estimate, axis=0),
                os.path.join(out_dir, output_name),
                masker,
                data_fn[0],
                command_str,
                use_bids=use_bids,
            )
        else:
            for te_idx in range(n_te):
                output_name = get_outname(
                    output_filename, f"echo-{te_idx + 1}_MAD", "nii.gz", use_bids
                )
                out_bids_keywords.append(f"echo-{te_idx + 1}_MAD")
                if te_idx == 0:
                    y_echo = data_masked[:n_scans, :]
                else:
                    y_echo = data_masked[te_idx * n_scans : (te_idx + 1) * n_scans, :]
                _, _, noise_estimate = select_lambda.select_lambda(hrf=hrf, y=y_echo)
                write_data(
                    np.expand_dims(noise_estimate, axis=0),
                    os.path.join(out_dir, output_name),
                    masker,
                    data_fn[0],
                    command_str,
                    use_bids=use_bids,
                )

        # Save lambda
        out_keyword = "lambda"
        if use_bids:
            out_keyword = f"stat-{out_keyword}_statmap"
        output_name = get_outname(output_filename, out_keyword, "nii.gz", use_bids)
        out_bids_keywords.append(out_keyword)
        write_data(
            np.expand_dims(lambda_map, axis=0),
            os.path.join(out_dir, output_name),
            masker,
            data_fn[0],
            command_str,
            use_bids=use_bids,
        )

    # Save BIDS compatible sidecar file
    if use_bids:
        write_json(out_bids_keywords, out_dir)

    LGR.info("Results saved.")

    LGR.info("pySPFM finished.")
    utils.teardown_loggers()


def _main():
    """pySPFM entry point"""
    command_str = "pySPFM " + " ".join(sys.argv[1:])
    options = _get_parser().parse_args()
    pySPFM(**vars(options), command_str=command_str)


if __name__ == "__main__":
    _main()



import numpy as np

from pySPFM.deconvolution import fista


def test_fista(sim_data, sim_hrf, fista_results):
    y = np.load(sim_data, allow_pickle=True)
    hrf_matrix = np.load(sim_hrf, allow_pickle=True)

    # Run FISTA without pylops
    estimates, lambda_ = fista.fista(hrf_matrix, y, group=0.2, use_pylops=False, max_iter=50)

    # Compare estimates
    assert np.allclose(lambda_, np.repeat(np.array([0.60157822]), y.shape[1], axis=0))
    assert np.allclose(estimates, np.load(fista_results, allow_pickle=True), atol=1e-6)


def test_fista_pylops(sim_data, sim_hrf, pylops_results):
    y = np.load(sim_data, allow_pickle=True)
    hrf_matrix = np.load(sim_hrf, allow_pickle=True)

    # Run FISTA with pylops
    estimates, lambda_ = fista.fista(hrf_matrix, y, group=0.2, use_pylops=True, max_iter=50)

    # Compare estimates
    assert np.allclose(lambda_, np.repeat(np.array([0.60157822]), y.shape[1], axis=0))
    assert np.allclose(estimates, np.load(pylops_results, allow_pickle=True), atol=1e-6)



"""
Functions for parsers.
"""
import argparse
import os.path as op


def check_hrf_value(string, is_parser=True):
    """
    Check if argument is 'spm' or 'glover',
    or is an existing '.1D' or '.txt' file.
    """
    if string == "spm" or string == "glover":
        return string
    elif (string.endswith(".1D") or string.endswith(".txt")) and is_parser:
        return is_valid_file(argparse.ArgumentParser(), string)
    else:
        raise argparse.ArgumentTypeError(
            "HRF model must be 'spm', 'glover', or a custom HRF file with the '.1D' "
            "or '.txt' extension."
        )


def is_valid_file(parser, arg):
    """
    Check if argument is existing file.
    """
    if not op.isfile(arg) and arg is not None:
        parser.error("The file {0} does not exist!".format(arg))

    return arg



import os.path as op

import nibabel as nib
import numpy as np
from nilearn.maskers import NiftiMasker

from pySPFM import io


def test_read_data(nih_five_echo_1, mask_five_echo):
    data, masker = io.read_data(nih_five_echo_1, mask_five_echo)
    assert data.shape == (75, 4)
    assert masker.mask_img.shape == (41, 52, 28)


def test_write_data(nih_five_echo_1, mask_five_echo, testpath):
    orig_img = nib.load(nih_five_echo_1)
    dummy_data = np.ones((75, 4))
    out_name = op.join(testpath, "test_write_data.nii.gz")

    masker = NiftiMasker(mask_img=mask_five_echo, standardize=False)
    masker.fit()

    io.write_data(dummy_data, out_name, masker, orig_img, "test_write_data")

    # Check that the file was created
    assert op.exists(out_name)



import argparse
import datetime
import logging
import os
import sys
from os import path as op

import nibabel as nib
import numpy as np

from pySPFM import __version__, utils
from pySPFM.deconvolution import debiasing, hrf_generator
from pySPFM.io import read_data, write_data, write_json
from pySPFM.utils import get_outname
from pySPFM.workflows.parser_utils import check_hrf_value, is_valid_file

LGR = logging.getLogger("GENERAL")
RefLGR = logging.getLogger("REFERENCES")


def _get_parser():
    """
    Parse command line inputs for this function.

    Returns
    -------
    parser.parse_args() : argparse dict

    Notes
    -----
    # Argument parser follow template provided by RalphyZ.
    # https://stackoverflow.com/a/43456577
    """
    parser = argparse.ArgumentParser()
    optional = parser._action_groups.pop()
    required = parser.add_argument_group("Required Argument:")
    required.add_argument(
        "-i",
        "--input",
        dest="data_fn",
        type=lambda x: is_valid_file(parser, x),
        nargs="+",
        help="The name of the file containing fMRI data.",
        required=True,
    )
    required.add_argument(
        "-a",
        "--auc",
        dest="auc_fn",
        type=lambda x: is_valid_file(parser, x),
        help="The name of the file containing AUC data.",
        required=True,
    )
    required.add_argument(
        "-m",
        "--mask",
        dest="mask_fn",
        nargs="+",
        type=lambda x: is_valid_file(parser, x),
        help=(
            "The name of the files containing the mask for the fMRI data and the AUC "
            "thresholding mask."
        ),
        required=True,
    )
    required.add_argument(
        "-o",
        "--output",
        dest="output_filename",
        type=str,
        help="The name of the output file with no extension.",
        required=True,
    )
    required.add_argument(
        "-tr",
        dest="tr",
        type=float,
        help="TR of the fMRI data acquisition.",
        required=True,
    )
    optional.add_argument(
        "-thr",
        "--threshold",
        dest="thr",
        help=(
            "Percentile to threshold the AUC data with or the threshold value itself."
            "The percentile is applied to the second mask provided with the '-m' flag if the "
            "second mask is a binary mask. If the second mask is not binary, the values on the "
            "second mask are used as the threshold. "
            "When the threshold value is given, the second mask is ignored. "
            "Percentiles are given in the range [1, 100], while threshold values are given in "
            "the range [0, 1). "
            "Default is 95."
        ),
        type=float,
        default=95.0,
    )
    optional.add_argument(
        "--strategy",
        dest="thr_strategy",
        help=(
            "Strategy to threshold the AUC data with. If the second mask is a binary mask, "
            "the can be applied with a static threshold ('static') or a time-dependet threshold "
            "('time'). Default is 'static'."
        ),
        type=str,
        default="static",
        choices=["static", "time"],
    )
    optional.add_argument(
        "-d",
        "--dir",
        dest="out_dir",
        type=str,
        help="Output directory. Default is current.",
        default=".",
    )
    optional.add_argument(
        "-te",
        dest="te",
        nargs="*",
        type=float,
        help="List with TE of the fMRI data acquisition.",
        default=[0],
    )
    optional.add_argument(
        "-hrf",
        "--hrf",
        dest="hrf_model",
        type=check_hrf_value,
        help=(
            "HRF model to use. Default is 'spm'. Options are 'spm', 'glover', or a custom HRF "
            "file with the '.1D' or '.txt' extension."
        ),
        default="spm",
    )
    optional.add_argument(
        "-block",
        "--block",
        dest="block_model",
        action="store_true",
        help="Estimate innovation signals. Default = False.",
        default=False,
    )
    optional.add_argument(
        "-jobs",
        "--jobs",
        dest="n_jobs",
        type=int,
        help="Number of jobs to parallelize for loops (default = 4).",
        default=4,
    )
    optional.add_argument(
        "-bids",
        "--bids",
        dest="use_bids",
        action="store_true",
        help=(
            "Use BIDS-style suffix on the given `output` (default = False). pySPFM assumes that "
            "`output` follows the BIDS convention. Not using this option will default to using "
            "AFNI to update the header of the output."
        ),
        default=False,
    )
    optional.add_argument(
        "--group",
        dest="group",
        action="store_true",
        help=(
            "Consider consecutive coefficients as belonging to the same block activation. Default "
            "= False."
        ),
        default=False,
    )
    optional.add_argument(
        "--group-distance",
        dest="group_distance",
        type=int,
        help=(
            "Maximum distance between coefficients to be considered part of the same block "
            "activation. Default = 3."
        ),
        default=3,
    )
    optional.add_argument(
        "-debug",
        "--debug",
        dest="debug",
        action="store_true",
        help="Logs in the terminal will have increased "
        "verbosity, and will also be written into "
        "a .tsv file in the output directory.",
        default=False,
    )
    optional.add_argument(
        "-quiet",
        "--quiet",
        dest="quiet",
        help=argparse.SUPPRESS,
        action="store_true",
        default=False,
    )
    optional.add_argument("-v", "--version", action="version", version=f"%(prog)s {__version__}")

    parser._action_groups.append(optional)

    return parser


def auc_to_estimates(
    data_fn,
    auc_fn,
    mask_fn,
    output_filename,
    tr,
    thr=95.0,
    thr_strategy="static",
    out_dir=".",
    te=[0],
    hrf_model="spm",
    block_model=False,
    n_jobs=4,
    use_bids=False,
    group=False,
    group_distance=3,
    debug=False,
    quiet=False,
    command_str=None,
):
    # Generate output directory if it doesn't exist
    out_dir = op.abspath(out_dir)
    if not op.isdir(out_dir):
        os.mkdir(out_dir)

    # Save command into sh file, if the command-line interface was used
    if command_str is not None:
        with open(os.path.join(out_dir, "call.sh"), "w") as command_file:
            command_file.write(command_str)
    LGR = logging.getLogger("GENERAL")

    # create logfile name
    basename = "auc_to_estimates_"
    extension = "tsv"
    start_time = datetime.datetime.now().strftime("%Y-%m-%dT%H%M%S")
    logname = op.join(out_dir, (basename + start_time + "." + extension))
    refname = op.join(out_dir, "_references.txt")
    utils.setup_loggers(logname, refname, quiet=quiet, debug=debug)

    LGR.info(f"Using output directory: {out_dir}")

    n_te = len(te)

    if all(i >= 1 for i in te):
        te = [x / 1000 for x in te]

    LGR.info("Reading data...")
    if n_te == 1:
        data_masked, masker = read_data(data_fn[0], mask_fn[0])
        n_scans = data_masked.shape[0]
    elif n_te > 1:
        # If the first element of data_fn has spaces, it is a list of paths
        # Convert it into a list
        if " " in data_fn[0]:
            data_fn = data_fn[0].split(" ")

        for te_idx in range(n_te):
            data_temp, masker = read_data(data_fn[te_idx], mask_fn[0])
            if te_idx == 0:
                data_masked = data_temp
                n_scans = data_temp.shape[0]
            else:
                data_masked = np.concatenate((data_masked, data_temp), axis=0)

            LGR.info(f"{te_idx + 1}/{n_te} echoes...")

    LGR.info("Data read.")

    LGR.info("Reading AUC data...")
    auc = read_data(auc_fn, mask_fn[0])[0]
    LGR.info("AUC data read.")

    # Threshold the AUC if thr is not 0 and mask_fn has two elements
    if thr != 0 and len(mask_fn) == 2:
        # Read the second mask
        auc_mask = nib.load(mask_fn[1])

        # If the mask is 3D, then it is a binary mask or a static threshold
        if len(auc_mask.shape) == 3:
            # If the mask is binary, then read the AUC values inside of the mask
            if np.max(auc_mask.get_fdata()) == 1:
                auc_thr_values = read_data(auc_fn, mask_fn[1])[0]

                if thr_strategy == "static":
                    LGR.info(
                        f"Thresholding AUC values with a {thr}th percentile static threshold..."
                    )
                    # Threshold the whole-brain AUC based on the thr percentile of the AUC values
                    # in the mask
                    auc_thr = auc - np.percentile(auc_thr_values, int(thr))
                    auc_thr[auc_thr < 0] = 0
                else:
                    LGR.info(
                        f"Thresholding AUC values with a {thr}th percentile time-dependet "
                        "threshold..."
                    )

                    # Calculate and apply percentile at each TR
                    auc_thr = np.zeros(auc.shape)
                    for tr_idx in range(n_scans):
                        auc_thr[tr_idx, :] = auc[tr_idx, :] - np.percentile(
                            auc_thr_values[tr_idx, :], int(thr)
                        )
                        auc_thr[tr_idx, auc_thr[tr_idx, :] < 0] = 0

            # If the mask is a static threshold, then apply it to the AUC values
            else:
                LGR.info("Thresholding AUC values based on the given 3D threshold...")
                auc_mask_data = masker.fit_transform(auc_mask)

                # Threshold the AUC values
                auc_thr = auc - auc_mask_data
                auc_thr[auc_thr < 0] = 0

        # If the mask is 4D, then it is a time-dependent threshold
        elif len(auc_mask.shape) == 4:
            LGR.info("Thresholding AUC values based on the given 4D threshold...")
            # Read the time-dependent threshold
            auc_mask_data = masker.fit_transform(auc_mask)

            # Threshold the AUC
            auc_thr = auc - auc_mask_data
            auc_thr[auc_thr < 0] = 0
        else:
            raise ValueError("The mask used to threshold the AUC must be 3D or 4D.")
    elif 0 < thr < 1:
        LGR.info(f"Thresholding AUC values with a threshold of {thr}...")
        # Threshold the whole-brain AUC based on the thr provided by the user
        auc_thr = auc - thr
        auc_thr[auc_thr < 0] = 0
    elif thr > 1 and len(mask_fn) != 1:
        # Raise error if thr is not 0 and mask_fn has only one element
        raise ValueError(
            "If the threshold is a percentile, then the 'mask' flag must have two elements."
        )
    # If thr is 0, then the AUC is supposed to be already thresholded
    else:
        LGR.warning("Threshold 0 selected. AUC is assumed to be already thresholded.")
        auc_thr = auc

    LGR.info("AUC data thresholded.")

    # Generate design matrix with shifted versions of HRF
    LGR.info("Generating design matrix with shifted versions of HRF...")
    hrf_obj = hrf_generator.HRFMatrix(te=te, block=block_model, model=hrf_model)
    hrf = hrf_obj.generate_hrf(tr=tr, n_scans=n_scans).hrf_

    # Solve ordinary least squares problem to calculate estimates
    LGR.info("Calculating estimates...")
    if block_model:
        estimates_spike = debiasing.debiasing_block(hrf, data_masked, auc_thr, n_jobs)
        fitts = np.dot(hrf, estimates_spike)
    else:
        estimates_spike, fitts = debiasing.debiasing_spike(
            hrf, data_masked, auc_thr, n_jobs, group=group, group_dist=group_distance
        )

    LGR.info("Estimates calculated.")

    # Save estimates and thresholded AUC
    LGR.info("Saving results...")
    out_bids_keywords = ["AUC"]

    output_name = get_outname(output_filename, "aucThresholded", "nii.gz", use_bids)
    write_data(
        auc,
        os.path.join(out_dir, output_name),
        masker,
        data_fn[0],
        command_str,
        use_bids,
    )

    # Save activity-inducing signal
    output_name = get_outname(output_filename, "activityInducing", "nii.gz", use_bids)
    out_bids_keywords.append("activityInducing")
    write_data(
        estimates_spike,
        os.path.join(out_dir, output_name),
        masker,
        data_fn[0],
        command_str,
        use_bids=use_bids,
    )

    # Save fitts
    if n_te == 1:
        output_name = get_outname(output_filename, "denoised_bold", "nii.gz", use_bids)
        out_bids_keywords.append("denoised_bold")
        write_data(
            fitts,
            os.path.join(out_dir, output_name),
            masker,
            data_fn[0],
            command_str,
            use_bids=use_bids,
        )
    elif n_te > 1:
        for te_idx in range(n_te):
            te_data = fitts[te_idx * n_scans : (te_idx + 1) * n_scans, :]
            output_name = get_outname(
                f"{output_filename}_echo-{te_idx + 1}", "denoised_bold", "nii.gz", use_bids
            )
            out_bids_keywords.append(f"echo-{te_idx + 1}_desc-denoised_bold")
            write_data(
                te_data,
                os.path.join(out_dir, output_name),
                masker,
                data_fn[0],
                command_str,
                use_bids=use_bids,
            )

    # Save BIDS compatible sidecar file
    if use_bids:
        write_json(out_bids_keywords, out_dir)

    LGR.info("Results saved.")

    LGR.info("auc_to_estimates finished.")
    utils.teardown_loggers()


def _main():
    """auc_to_estimates entry point"""
    command_str = "auc_to_estimates " + " ".join(sys.argv[1:])
    options = _get_parser().parse_args()
    auc_to_estimates(**vars(options), command_str=command_str)


if __name__ == "__main__":
    _main()



import os
import ssl
from urllib.request import urlretrieve

import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--skipintegration", action="store_true", default=False, help="Skip integration tests."
    )


@pytest.fixture
def skip_integration(request):
    return request.config.getoption("--skipintegration")


def fetch_file(osf_id, path, filename):
    """
    Fetches file located on OSF and downloads to `path`/`filename`1
    Parameters
    ----------
    osf_id : str
        Unique OSF ID for file to be downloaded. Will be inserted into relevant
        location in URL: https://osf.io/{osf_id}/download
    path : str
        Path to which `filename` should be downloaded. Ideally a temporary
        directory
    filename : str
        Name of file to be downloaded (does not necessarily have to match name
        of file on OSF)
    Returns
    -------
    full_path : str
        Full path to downloaded `filename`
    """
    # This restores the same behavior as before.
    # this three lines make tests dowloads work in windows
    if os.name == "nt":
        orig_sslsocket_init = ssl.SSLSocket.__init__
        ssl.SSLSocket.__init__ = (
            lambda *args, cert_reqs=ssl.CERT_NONE, **kwargs: orig_sslsocket_init(
                *args, cert_reqs=ssl.CERT_NONE, **kwargs
            )
        )
        ssl._create_default_https_context = ssl._create_unverified_context
    url = "https://osf.io/{}/download".format(osf_id)
    full_path = os.path.join(path, filename)
    if not os.path.isfile(full_path):
        urlretrieve(url, full_path)
    return full_path


@pytest.fixture(scope="session")
def testpath(tmp_path_factory):
    """Test path that will be used to download all files"""
    return tmp_path_factory.getbasetemp()


@pytest.fixture
def pylops_results(testpath):
    return fetch_file("bmhtr", testpath, "pylops_fista.npy")


@pytest.fixture
def sim_data(testpath):
    return fetch_file("789z6", testpath, "sim_data.npy")


@pytest.fixture
def sim_hrf(testpath):
    return fetch_file("j2u6n", testpath, "sim_hrf.npy")


@pytest.fixture
def mask_five_echo(testpath):
    return fetch_file("jv5fn", testpath, "mask_five_echo.nii.gz")


@pytest.fixture
def spm_single_echo_block(testpath):
    return fetch_file("4k85j", testpath, "spm_single_echo_block.npy")


@pytest.fixture
def spm_single_echo(testpath):
    return fetch_file("zt3pm", testpath, "spm_single_echo.npy")


@pytest.fixture
def glover_multi_echo(testpath):
    return fetch_file("xke79", testpath, "glover_multi_echo.npy")


@pytest.fixture
def nih_five_echo_1(testpath):
    return fetch_file("em9r8", testpath, "p06.SBJ01_S09_Task11_e1.sm.nii.gz")


@pytest.fixture
def fista_results(testpath):
    return fetch_file("3a925", testpath, "fista_results.npy")


@pytest.fixture
def coef_path_results(testpath):
    return fetch_file("nxgeq", testpath, "coef_path.npy")


@pytest.fixture
def test_AUC(testpath):
    return fetch_file("spzu9", testpath, "test_AUC.nii.gz")


@pytest.fixture
def mean_AUC(testpath):
    return fetch_file("vhdm9", testpath, "mean_AUC.nii.gz")


@pytest.fixture
def auc_4D_thr(testpath):
    return fetch_file("2pqmy", testpath, "test_AUC_4D_thr.nii.gz")



import numpy as np

from pySPFM.deconvolution import debiasing


def test_debiasing_spike(sim_data, sim_hrf):
    sim_data = np.load(sim_data, allow_pickle=True)[:, 0]
    sim_data = sim_data[:, np.newaxis]
    sim_hrf = np.load(sim_hrf, allow_pickle=True)

    nt = sim_data.shape[0]

    from pySPFM.deconvolution.hrf_generator import HRFMatrix

    hrf_generator = HRFMatrix(block=False, te=[0.018, 0.036, 0.052])
    sim_hrf = hrf_generator.generate_hrf(1, sim_data.shape[0]).hrf_

    sim_data = np.repeat(sim_data, 3, axis=0)

    estimates = np.zeros((nt, 1))
    estimates_idxs = [
        35,
        50,
        65,
        80,
        151,
        152,
        153,
        154,
        155,
        156,
        157,
        158,
        159,
        160,
        210,
        211,
        212,
        213,
        214,
    ]
    estimates[estimates_idxs] = 1

    beta, fitts = debiasing.debiasing_spike(
        sim_hrf,
        sim_data,
        estimates,
        group=True,
    )
    # Check that beta is different from zero on the same indices as estimates
    assert np.all(beta[estimates != 0] != 0)

    # Check that the convolution of the hrf with estimates is different from zero on the same
    # indices as fitts
    assert np.all(fitts[np.dot(sim_hrf, estimates) != 0] != 0)


def test_debiasing_block(sim_data, sim_hrf):
    sim_data = np.load(sim_data, allow_pickle=True)[:, 0]
    sim_data = sim_data[:, np.newaxis]
    sim_hrf = np.load(sim_hrf, allow_pickle=True)

    nt = sim_data.shape[0]
    sim_hrf = np.dot(sim_hrf, np.tril(np.ones(nt)))

    estimates = np.zeros((nt, 1))
    estimates[35] = 1
    estimates[50] = 1
    estimates[65] = 1
    estimates[80] = 1
    estimates[151:161] = 1
    estimates[210:215] = 1

    beta = debiasing.debiasing_block(sim_hrf, sim_data, estimates_matrix=estimates)

    # Check that the derivative of beta is equal to estimates
    assert np.allclose(np.diff(beta) != 0, estimates != 0)



"""Least Angle Regression (LARS) method for deconvolution."""
import logging

import numpy as np
from sklearn.linear_model import lars_path

LGR = logging.getLogger("GENERAL")


def select_optimal_lambda(residuals, non_zero_count, n_scans, criterion="bic"):
    """Select optimal lambda based on the model selection criterion (BIC and AIC)

    Parameters
    ----------
    residuals : ndarray
        Residuals of the model
    non_zero_count : ndarray
        Number of non-zero coefficients for each lambda
    n_scans : int
        Number of scans
    criterion : str, optional
        Criterion to find the optimal solution, by default "bic"

    Returns
    -------
    index_optimal_lambda : int
        Index of the optimal lambda
    """
    if criterion == "bic":
        # BIC regularization curve
        optimization_curve = n_scans * np.log(residuals) + np.log(n_scans) * non_zero_count
    elif criterion == "aic":
        # AIC regularization curve
        optimization_curve = n_scans * np.log(residuals) + 2 * non_zero_count

    # Optimal lambda is given by the minimum of the optimization curve
    idx_optimal_lambda = np.argmin(optimization_curve)

    return idx_optimal_lambda


def solve_regularization_path(X, y, n_lambdas, criterion="bic"):
    """Solve the regularization path with the LARS algorithm.

    Parameters
    ----------
    X : ndarray
        Design matrix
    y : ndarray
        Voxel time-series
    n_lambdas : int
        Number of lambdas to be tested
    criterion : str, optional
        Criterion to find the optimal solution, by default "bic"

    Returns
    -------
    coef_path : ndarray
        Estimates of the coefficients for the optimal lambda
    lambdas : ndarray
        Lambda of the optimal solution
    """
    n_scans = X.shape[1]

    # If y is a vector, add a dimension to make it a matrix
    if y.ndim == 1:
        y = y[:, np.newaxis]

    # Initialize variables to store the results
    coef_path = np.zeros((n_scans, n_lambdas))
    lambdas = np.zeros((n_lambdas,))

    # LARS path
    lambdas_temp, _, coef_path_temp = lars_path(
        X,
        np.squeeze(y),
        method="lasso",
        Gram=np.dot(X.T, X),
        Xy=np.dot(X.T, np.squeeze(y)),
        max_iter=n_lambdas - 1,
        eps=1e-9,
    )

    # Store the results
    coef_path[:, : len(lambdas_temp)] = coef_path_temp
    lambdas[: len(lambdas_temp)] = lambdas_temp

    # Compute residuals for model selection criterion (BIC and AIC)
    residuals = np.sum((np.repeat(y, n_lambdas, axis=-1) - np.dot(X, coef_path)) ** 2, axis=0)

    if criterion == "stability":
        optimal_lambda = lambdas
        coefs = coef_path
    else:
        optimal_lambda_idx = select_optimal_lambda(
            residuals, np.count_nonzero(coef_path, axis=0), n_scans, criterion
        )
        optimal_lambda = lambdas[optimal_lambda_idx]
        coefs = coef_path[:, optimal_lambda_idx]

    return coefs, optimal_lambda



import logging
import os

import jax
import jax.numpy as jnp
import numpy as np

from pySPFM.deconvolution.lars import solve_regularization_path

LGR = logging.getLogger("GENERAL")


def get_subsampling_indices(n_scans, n_echos, mode="same"):
    if "mode" in os.environ.keys():  # only for testing
        np.random.seed(200)
    # Subsampling for Stability Selection
    if mode == "different":  # different time points are selected across echoes
        subsample_idx = np.sort(
            np.random.choice(range(n_scans), int(0.6 * n_scans), 0)
        )  # 60% of timepoints are kept
        for i in range(n_echos - 1):
            subsample_idx = np.concatenate(
                (
                    subsample_idx,
                    np.sort(
                        np.random.choice(
                            range((i + 1) * n_scans, (i + 2) * n_scans),
                            int(0.6 * n_scans),
                            0,
                        )
                    ),
                )
            )
    elif mode == "same":  # same time points are selected across echoes
        subsample_idx = np.sort(
            np.random.choice(range(n_scans), int(0.6 * n_scans), 0)
        )  # 60% of timepoints are kept

    return subsample_idx


def calculate_auc(coefs, lambdas, n_surrogates):
    """Calculate the AUC for a TR.

    Parameters
    ----------
    coefs : np.ndarray
        Matrix of coefficients of shape (n_lambdas, n_surrogates).
    lambdas : np.ndarray
        Array of lambdas of shape (n_lambdas).
    n_surrogates : int
        Number of surrogates.

    Returns
    -------
    auc : float
        AUC for a TR.
    """

    # Sum of all lambdas
    lambdas_sum = jnp.sum(lambdas)

    # Binarize coefficients
    coefs = jnp.where(coefs != 0, 1, 0)
    sum_ = 0

    # If coefs is two-dimensional, use the first dimension to calculate the AUC
    if coefs.ndim == 2:
        probs = jnp.sum(coefs, axis=1) / n_surrogates
        for i in range(len(lambdas)):
            sum_ += probs[i] * lambdas[i] / lambdas_sum

    # If coefs is one-dimensional, use the whole array to calculate the AUC
    elif coefs.ndim == 1:
        for i in range(len(lambdas)):
            sum_ += coefs[i] * lambdas[i] / lambdas_sum

    return sum_


def _get_tr_lambdas(estimates, lambdas, n_lambdas, n_surrogates):
    # Check if all lambdas are the same across axis 1.
    # If they are not, merge all lambdas into single, shared space.
    if not jnp.allclose(jnp.sum(lambdas, axis=1), n_lambdas * lambdas[:, 0]):
        estimates_tr, lambdas_tr = _generate_shared_lambdas_space(
            estimates, lambdas, n_lambdas, n_surrogates
        )
    else:
        estimates_tr = estimates
        lambdas_tr = lambdas[:, 0]

    return estimates_tr, lambdas_tr


def _generate_shared_lambdas_space(coefs, lambdas, n_lambdas, n_surrogates):
    """Generate shared space of lambdas and coefficients.

    Parameters
    ----------
    coefs : np.ndarray
        Coefficients of shape (n_lambdas, n_surrogates).
    lambdas : np.ndarray
        Lambdas of shape (n_lambdas, n_surrogates).
    n_lambdas : int
        Number of lambdas.
    n_surrogates : int
        Number of surrogates.

    Returns
    -------
    coefs_sorted : np.ndarray
        Sorted coefficients.
    lambdas_sorted : np.ndarray
        Sorted lambdas.
    """
    # Create shared space of lambdas and coefficients
    lambdas_shared = lambdas.reshape((n_lambdas * n_surrogates))
    coefs_shared = np.zeros((coefs.shape[0] * coefs.shape[1]))

    # Project lambdas and coefficients into shared space
    for i in range(lambdas.shape[0]):
        coefs_shared[i * coefs.shape[1] : (i + 1) * coefs.shape[1]] = np.squeeze(coefs[i, :])

    # Sort lambdas and get the indices
    lambdas_sorted_idx = np.argsort(-lambdas_shared)
    lambdas_sorted = -np.sort(-lambdas_shared)

    # Sort coefficients
    coefs_sorted = coefs_shared[lambdas_sorted_idx]

    return coefs_sorted, lambdas_sorted


def stability_selection(hrf_norm, data, n_lambdas, n_surrogates):
    # Get n_scans, n_echos, n_voxels
    n_scans = hrf_norm.shape[1]
    n_echos = int(np.ceil(hrf_norm.shape[0] / n_scans))

    # Initialize variables to store the results
    estimates = np.zeros((n_scans, n_lambdas, n_surrogates))
    lambdas = np.zeros((n_lambdas, n_surrogates))

    # Generate surrogates and compute the regularization path
    stability_estimates = []
    for _ in range(n_surrogates):
        # Subsampling for Stability Selection
        subsample_idx = get_subsampling_indices(n_scans, n_echos)

        # Solve LARS
        fut_stability = solve_regularization_path(
            hrf_norm[subsample_idx, :], data[subsample_idx], n_lambdas, "stability"
        )
        stability_estimates.append(fut_stability)

    for surr_idx in range(n_surrogates):
        estimates[:, :, surr_idx] = np.squeeze(stability_estimates[surr_idx][0])
        lambdas[:, surr_idx] = np.squeeze(stability_estimates[surr_idx][1])

    # Calculate the AUC for each TR
    auc = np.zeros((n_scans))
    calculate_auc_jit = jax.jit(calculate_auc)
    for tr_idx in range(n_scans):
        # Get lambdas and coefficients for the TR
        estimates_tr, lambdas_tr = _get_tr_lambdas(
            estimates[tr_idx, :, :], lambdas, n_lambdas, n_surrogates
        )

        # Calculate AUC
        auc[tr_idx] = calculate_auc_jit(estimates_tr, lambdas_tr, n_surrogates).block_until_ready()

    return auc



"""Debiasing functions for PFM."""
import logging

import numpy as np
import scipy as sci
from dask import compute
from dask import delayed as delayed_dask
from scipy.signal import find_peaks
from sklearn.linear_model import RidgeCV

from pySPFM.utils import dask_scheduler

LGR = logging.getLogger("GENERAL")
RefLGR = logging.getLogger("REFERENCES")


def group_hrf(hrf, non_zero_idxs, group_dist=3):
    """Group HRFs based on the distance between non-zero coefficients.

    Parameters
    ----------
    hrf : (E x T) ndarray
        Matrix containing shifted HRFs in its columns. E stands for the number of volumes times
        the number of echo-times.
    non_zero_idxs : (T) ndarray
        Array containing the indexes of the non-zero coefficients.
    group_dist : int, optional
        Maximum distance between non-zero coefficients to be considered as part of the same group,
        by default 3

    Returns
    -------
    hrf_out : (E x T) ndarray
        Matrix containing shifted HRFs in its columns. E stands for the number of volumes times
        the number of echo-times.
    new_idxs : (T) ndarray
        Array containing the indexes of the non-zero coefficients.
    """
    temp = np.zeros(hrf.shape[0])
    hrf_out = np.zeros(hrf.shape)
    non_zeros_flipped = np.flip(non_zero_idxs)
    new_idxs = []

    for iter_idx, non_zero_idx in enumerate(non_zeros_flipped):
        if (
            iter_idx != len(non_zeros_flipped) - 1
            and abs(non_zero_idx - non_zeros_flipped[iter_idx + 1]) <= group_dist
        ):
            temp += hrf[:, non_zero_idx]
        else:
            temp += hrf[:, non_zero_idx]
            hrf_out[:, non_zero_idx] = temp
            temp = np.zeros(hrf.shape[0])
            new_idxs.append(non_zero_idx)

    new_idxs = np.flip(new_idxs)
    hrf_out = hrf_out[:, new_idxs]

    return hrf_out, new_idxs


def group_betas(betas, non_zero_idxs, group_dist=3):
    """Group betas based on the distance between non-zero coefficients.

    Parameters
    ----------
    betas : (T) ndarray
        Array containing the non-zero coefficients selected as neuronal-related.
    non_zero_idxs : (T) ndarray
        Array containing the indexes of the non-zero coefficients.
    group_dist : int, optional
        Maximum distance between non-zero coefficients to be considered as part of the same group,
        by default 3

    Returns
    -------
    betas : (T) ndarray
        Array containing the non-zero coefficients selected as neuronal-related.
    """
    for i in range(len(non_zero_idxs)):
        if i > 0 and (non_zero_idxs[i] - non_zero_idxs[i - 1]) <= group_dist:
            betas[non_zero_idxs[i]] = betas[non_zero_idxs[i - 1]]

    return betas


# Performs the debiasing step on an estimates_matrix timeseries obtained considering the
# integrator model
def innovation_to_block(hrf, y, estimates_matrix, is_ls):
    """Perform debiasing with the block model.

    Parameters
    ----------
    hrf : (E x T) ndarray
        Matrix containing shifted HRFs in its columns. E stands for the number of volumes times
        the number of echo-times.
    y : (T x S) ndarray
        Matrix with fMRI data provided to pySPFM.
    estimates_matrix : (T x S) ndarray
        Matrix containing the non-zero coefficients selected as neuronal-related.
    is_ls : bool
        Whether least squares is solved in favor of ridge regression.

    Returns
    -------
    beta : (T x S) ndarray
        Debiased activity-inducing signal obtained from estimated innovation signal.
    S : (T x L) ndarray
        Transformation matrix used to integrate the innovation signal into activity-inducing
        signal. L stands for the number of steps to integrate.
    """
    # Find indexes of nonzero coefficients
    nonzero_idxs = np.where(estimates_matrix != 0)[0]
    n_nonzero = len(nonzero_idxs)  # Number of nonzero coefficients

    # Initiates beta
    beta = np.zeros((estimates_matrix.shape))
    S = 0

    if n_nonzero != 0:
        # Initiates matrix S and array of labels
        S = np.zeros((hrf.shape[1], n_nonzero + 1))
        labels = np.zeros((estimates_matrix.shape[0]))

        # Gives values to S design matrix based on nonzeros in estimates_matrix
        # It also stores the labels of the changes in the design matrix
        # to later generate the debiased timeseries with the obtained betas
        for idx in range(n_nonzero + 1):
            if idx == 0:
                S[0 : nonzero_idxs[idx], idx] = 1
                labels[: nonzero_idxs[idx]] = idx
            elif idx == n_nonzero:
                S[nonzero_idxs[idx - 1] :, idx] = 1
                labels[nonzero_idxs[idx - 1] :] = idx
            else:
                S[nonzero_idxs[idx - 1] : nonzero_idxs[idx], idx] = 1
                labels[nonzero_idxs[idx - 1] : nonzero_idxs[idx]] = idx

        # Performs the least squares to obtain the beta amplitudes
        if is_ls:
            beta_amplitudes, _, _, _ = np.linalg.lstsq(
                np.dot(hrf, S), y, rcond=None
            )  # b-ax --> returns x
        else:
            clf = RidgeCV(alphas=[1e-3, 1e-2, 1e-1, 1, 10]).fit(np.dot(hrf, S), y)
            beta_amplitudes = clf.coef_

        # Positions beta amplitudes in the entire timeseries
        for amp_change in range(n_nonzero + 1):
            beta[labels == amp_change] = beta_amplitudes[amp_change]

    return beta, S


def do_debias_block(hrf, y, estimates_matrix, dist=2):
    """Perform debiasing with the block model.

    Parameters
    ----------
    hrf : (E x T) ndarray
        Matrix containing shifted HRFs in its columns. E stands for the number of volumes times
        the number of echo-times.
    y : (T x 1) ndarray
        Matrix with fMRI data provided to pySPFM in a voxel.
    estimates_matrix : (T x 1) ndarray
        Matrix containing the non-zero coefficients selected as neuronal-related in a voxel.
    dist : int, optional
        Minimum number of TRs in between of the peaks found, by default 2

    Returns
    -------
    beta_out : ndarray
        Debiased activity-inducing signal obtained from estimated innovation signal in a voxel.
    """
    # Keep only maximum values in estimates_matrix peaks
    temp = np.zeros((estimates_matrix.shape[0],))
    peak_idxs, _ = find_peaks(abs(estimates_matrix), distance=dist)
    temp[peak_idxs] = estimates_matrix[peak_idxs].copy()

    estimates_matrix = temp.copy()

    beta_out, _ = innovation_to_block(hrf, y, estimates_matrix, is_ls=True)

    return beta_out


def debiasing_block(hrf, y, estimates_matrix, n_jobs=4):
    """Voxelwise block model debiasing workflow.

    Parameters
    ----------
    hrf : (E x T) ndarray
        Matrix containing shifted HRFs in its columns. E stands for the number of volumes times
        the number of echo-times.
    y : (T x S) ndarray
        Matrix with fMRI data provided to pySPFM.
    estimates_matrix : (T x S) ndarray
        Matrix containing the non-zero coefficients selected as neuronal-related.
    n_jobs : int, optional
        Number of jobs to run in parallel, by default 4

    Returns
    -------
    beta_out : ndarray
        Debiased activity-inducing signal obtained from estimated innovation signal.
    """
    n_scans = estimates_matrix.shape[0]
    n_voxels = estimates_matrix.shape[1]

    # Initiates beta matrix
    beta_out = np.zeros((n_scans, n_voxels))

    LGR.info("Starting debiasing step...")
    # Performs debiasing
    dask_scheduler(n_jobs)
    futures = []
    for voxidx in range(n_voxels):
        fut = delayed_dask(do_debias_block, pure=False)(
            hrf, y[:, voxidx], estimates_matrix[:, voxidx]
        )
        futures.append(fut)
    debiased = compute(futures)[0]

    for vox_idx in range(n_voxels):
        beta_out[:, vox_idx] = debiased[vox_idx]

    LGR.info("Debiasing step finished")
    return beta_out


def do_debias_spike(hrf, y, estimates_matrix, group=False, group_dist=3):
    """Perform debiasing with the spike model.

    Parameters
    ----------
    hrf : (E x T) ndarray
        Matrix containing shifted HRFs in its columns. E stands for the number of volumes times
        the number of echo-times.
    y : (T x 1) ndarray
        Array with fMRI data of a voxel provided to pySPFM.
    estimates_matrix : (T x 1) ndarray
        Array containing the non-zero coefficients selected as neuronal-related.
    group : bool, optional
        Whether to group the spikes or not, by default False
    group_dist : int, optional
        Minimum number of TRs in between of the peaks found, by default 3

    Returns
    -------
    beta_out : ndarray
        Debiased activity-inducing signal in a voxel.
    fitts_out : ndarray
        Debiased activity-inducing signal convolved with the HRF in a voxel.
    """
    index_events_opt = np.where(abs(estimates_matrix) > 10 * np.finfo(float).eps)[0]
    beta2save = np.zeros((estimates_matrix.shape[0], 1))

    if group:
        hrf_events, index_events_opt_group = group_hrf(hrf, index_events_opt, group_dist)
    else:
        hrf_events = hrf[:, index_events_opt]

    coef_LSfitdebias, _, _, _ = sci.linalg.lstsq(hrf_events, y.T, cond=None)
    if group:
        beta2save[index_events_opt_group, 0] = coef_LSfitdebias
    else:
        beta2save[index_events_opt, 0] = coef_LSfitdebias
    fitts_out = np.squeeze(np.dot(hrf, beta2save))

    beta_out = beta2save.reshape(len(beta2save))
    if group:
        beta_out = group_betas(beta_out, index_events_opt, group_dist)

    return beta_out, fitts_out


def debiasing_spike(hrf, y, estimates_matrix, n_jobs=0, group=False, group_dist=3):
    """Perform voxelwise debiasing with spike model.

    Parameters
    ----------
    hrf : (E x T) ndarray
        Matrix containing shifted HRFs in its columns. E stands for the number of volumes times
        the number of echo-times.
    y : (T x S) ndarray
        Matrix with fMRI data provided to pySPFM.
    estimates_matrix : (T x S) ndarray
        Matrix containing the non-zero coefficients selected as neuronal-related.
    n_jobs : int, optional
        Number of jobs to run in parallel, by default 0
    group : bool, optional
        Whether to group the spikes or not, by default False
    group_dist : int, optional
        Minimum number of TRs in between of the peaks found, by default 3

    Returns
    -------
    beta_out : ndarray
        Debiased activity-inducing signal.
    fitts_out : ndarray
        Debiased activity-inducing signal convolved with the HRF.
    """
    beta_out = np.zeros(estimates_matrix.shape)
    fitts_out = np.zeros(y.shape)

    index_voxels = np.unique(np.where(abs(estimates_matrix) > 10 * np.finfo(float).eps)[1])

    LGR.info("Performing debiasing step...")
    _, cluster = dask_scheduler(n_jobs)
    futures = []
    for voxidx in range(len(index_voxels)):
        fut = delayed_dask(do_debias_spike, pure=False)(
            hrf,
            y[:, index_voxels[voxidx]],
            estimates_matrix[:, index_voxels[voxidx]],
            group,
            group_dist,
        )
        futures.append(fut)
    debiased = compute(futures)[0]

    for voxidx in range(len(index_voxels)):
        beta_out[:, index_voxels[voxidx]] = debiased[voxidx][0]
        fitts_out[:, index_voxels[voxidx]] = debiased[voxidx][1]

    LGR.info("Debiasing step finished")
    return beta_out, fitts_out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""pySPFM setup script"""
from setuptools import setup

import versioneer

if __name__ == "__main__":
    setup(
        name="pySPFM",
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        zip_safe=False,
    )



import numpy as np
import pytest

from pySPFM.deconvolution.spatial_regularization import clip, generate_delta


def test_clip():
    # Test clipping the input to the atlas
    input_array = np.array([1, 2, 3, 4, 5])
    atlas_array = np.array([1, 1, 1, 2, 2])

    clipped_array = clip(input_array, atlas_array)

    assert np.allclose(
        clipped_array, np.array([0.26726124, 0.53452248, 0.80178373, 0.62469505, 0.78086881])
    )


def test_generate_delta():
    # Test delta generation with dim=2
    h = generate_delta(dim=2)
    assert h.shape == (3, 3)

    # Test delta generation with dim=3
    h = generate_delta(dim=3)
    assert h.shape == (3, 3, 3)

    # Test delta generation with dim different from 2 or 3
    pytest.raises(ValueError, generate_delta, dim=1)



"""
Stub file for a guaranteed safe import of duecredit constructs:  if duecredit
is not available.
To use it, place it into your project codebase to be imported, e.g. copy as
    cp stub.py /path/tomodule/module/due.py
Note that it might be better to avoid naming it duecredit.py to avoid shadowing
installed duecredit.
Then use in your code as
    from .due import due, Doi, BibTeX
See  https://github.com/duecredit/duecredit/blob/master/README.md for examples.
Origin:     Originally a part of the duecredit
Copyright:  2015-2016  DueCredit developers
License:    BSD-2
"""

__version__ = "0.0.5"


class InactiveDueCreditCollector(object):
    """Just a stub at the Collector which would not do anything"""

    def _donothing(self, *args, **kwargs):
        """Perform no good and no bad"""
        pass

    def dcite(self, *args, **kwargs):
        """If I could cite I would"""

        def nondecorating_decorator(func):
            return func

        return nondecorating_decorator

    cite = load = add = _donothing

    def __repr__(self):
        return self.__class__.__name__ + "()"


def _donothing_func(*args, **kwargs):
    """Perform no good and no bad"""
    pass


try:
    from duecredit import BibTeX, Doi, Url, due

    if "due" in locals() and not hasattr(due, "cite"):
        raise RuntimeError("Imported due lacks .cite. DueCredit is now disabled")
except Exception as e:
    if type(e).__name__ != "ImportError":
        import logging

        logging.getLogger("duecredit").error("Failed to import duecredit due to %s" % str(e))
    # Initiate due stub
    due = InactiveDueCreditCollector()
    BibTeX = Doi = Url = _donothing_func

# Emacs mode definitions
# Local Variables:
# mode: python
# py-indent-offset: 4
# tab-width: 4
# indent-tabs-mode: nil
# End:



"""I/O."""
import json
import logging
import os.path as op
from subprocess import run

import nibabel as nib
from nilearn.image import new_img_like
from nilearn.maskers import NiftiLabelsMasker, NiftiMasker

from pySPFM.utils import get_keyword_description

LGR = logging.getLogger("GENERAL")


def read_data(data_fn, mask_fn):
    """Read data from filename and apply mask.

    Parameters
    ----------
    data_fn : str or path
        Path to data to be read.
    mask_fn : str or path
        Path to mask to be applied.

    Returns
    -------
    data : (T x S) ndarray
        Data in 2D.
    masker : nilearn.maskers.NiftiMasker
        Masker.
    """
    # Read data
    data_img = nib.load(data_fn)

    # Load mask and calculate maximum value
    mask_img = nib.load(mask_fn)
    mask_max = mask_img.get_fdata().max()

    # Check if mask is binary
    if mask_max > 1:
        LGR.info("Masking data with atlas.")
        masker = NiftiLabelsMasker(labels_img=mask_img, standardize=False, strategy="mean")
    elif mask_max == 1:
        LGR.info("Masking data with binary mask.")
        masker = NiftiMasker(mask_img=mask_img, standardize=False)
    else:
        raise ValueError("Mask is not binary or an atlas.")

    data = masker.fit_transform(data_img)

    return data, masker


def update_header(filename, command):
    """Update history of data to be read with 3dInfo.

    Parameters
    ----------
    filename : str or path
        Path to the file that is getting the header updated.
    command : str
        pySPFM command to add to the header.
    """
    run(f"3dcopy {filename} {filename} -overwrite", shell=True)
    run(f'3dNotes -h "{command}" {filename}', shell=True)


def write_data(data, filename, masker, orig_img, command, use_bids=False):
    """Write data into NIFTI file.

    Parameters
    ----------
    data : (T x S)
        Data in 2D.
    filename : str or path
        Name of the output file.
    masker : nilearn.maskers.NiftiMasker
        Masker.
    orig_img : nibabel.nifti1.Nifti1Image or str or path
        Original data.
    command : str
        pySPFM command to add to the header.
    use_bids : bool, optional
        Whether to use BIDS format, by default False
    """
    # Only copy header if it's going to get updated by AFNI
    copy_header = False
    if not use_bids:
        copy_header = True

    # If orig_img is a string, load it
    if isinstance(orig_img, str):
        orig_img = nib.load(orig_img)

    # Transform data back to 4D, generate new image and save it
    out_img = masker.inverse_transform(data)
    new_img = new_img_like(
        orig_img, out_img.get_fdata(), affine=orig_img.affine, copy_header=copy_header
    )
    new_img.to_filename(filename)

    # Update header with AFNI if BIDS is not required
    if not use_bids:
        update_header(filename, command)


def write_json(keywords, out_dir):
    """Write dataset description into JSON file.

    Parameters
    ----------
    keywords : list
        List of keywords to be added to the JSON file.
    out_dir : str or path
        Path to the output directory.
    """
    # Create dictionary with all the information
    out_dict = {}

    # Iterate over all the keywords and add their description and method to the dictionary
    for keyword in keywords:
        out_dict[keyword] = {}
        out_dict[keyword]["description"] = get_keyword_description(keyword)
        out_dict[keyword]["method"] = "pySPFM"

        # Add units
        if "bold" in keyword:
            out_dict[keyword]["units"] = "percent"
        elif "activityInducing" in keyword:
            # Count how many keywords have "bold" in their name
            bold_count = 0
            for k in keywords:
                if "bold" in k:
                    bold_count += 1
            if bold_count > 1:
                out_dict[keyword]["units"] = "s-1"

    # Create output filename
    outname = "dataset_description.json"

    # Write json file
    with open(op.join(out_dir, outname), "w") as f:
        f.write(json.dumps(out_dict, indent=4))
